/*
 * m_SegSVM_CCCP.cpp
 *
 * Created on: Aug 1, 2009
 * Author: Minh Hoai Nguyen, Carnegie Mellon University
 * Email:  minhhoai@cmu.edu, minhhoai@gmail.com
 */

#ifndef M_SEGSVM_CCCP
#define M_SEGSVM_CCCP

#include "m_utils.h"
#include "mex.h"
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ilcplex/ilocplex.h>

using namespace std;

class SegMCSVM_OptRslt{
public:
   double *w, **objVals, **wDeviation;
   double *all_ws; // history of w's.
   int nIter1, *nCGIters;
};

/*
 * Segment based MCSVM, optimization with Concave Convex Constrain Programming
 * \param featGen: a feature generation object, that supports three methods:
 *    + getNegFeats
 *    + getPosFeats
 * \param C: C parameter of SVM
 * \param nBin: # of dim of the weight vector (# of histogram bins)
 * \param nClass: # of classes
 * \param nTr: # of training samples for each class
 * \param runIDString: some string to be added to printing messages for discrimination
 */
template<class Feat>
SegMCSVM_OptRslt SegMCSVM_CCCP(Feat featGen, int *lbs, double C, int nBin, int nClass, int nTr, char *runIDString){
   /*Internal Parameters*/
   int MAX_ITERS_1 = 50; // no of iters for coordinate descent
   int MAX_ITERS_2 = 100;// no of iters for constraint generation
   int INIT_CONSTR_PER_SAMPLE = 10; // no of initial constraints for each training sample
   int VERBOSE = 1;
   double EXTRA_VIOLATION_TOLFAC = 0.000001; // stopping condition for extra violation (inner loop)
   double OBJ_REDUCTION_TOLFAC   = 0.00001; // stopping condition for obj. reduction (outer loop)

   double deviation;
   clock_t curTime   = time(NULL);
   clock_t startTime = curTime;

   IloEnv env;
   SegMCSVM_OptRslt optRslt;
   try{
      mexPrintf("SegMCSVM_CCCP: initializing and setting variables\n");
      IloModel mod(env);
      IloCplex cplex(mod); // initialize Cplex

      //variables
      int wSz = nBin*nClass;
      IloNumVarArray xw(env, wSz, -IloInfinity, IloInfinity);
      IloNumVarArray xxi(env, nTr, 0.0, IloInfinity);

      //associated values of the above variables
      IloNumArray w(env);
      IloNumArray xi(env);

      for (int i = 0; i < wSz; i++) w.add(0);
      for (int i = 0; i < nTr; i++) xi.add(0);

      mexPrintf("Cplex: adding the objective\n");
      IloExpr objExp(env);
      for (int i=0; i < wSz; i++) objExp += (1.0/nClass)*xw[i]*xw[i];
      for (int i=0; i < nTr; i++) objExp += (C/nTr)*xxi[i];
      mod.add(IloMinimize(env, objExp));
      objExp.end();

      // valid entry: CPX_ALG_AUTOMATIC, CPX_ALG_PRIMAL, CPX_ALG_DUAL, CPX_ALG_BARRIER
      int solverAlgo = CPX_ALG_DUAL;
      int solverVerbose = 0; //0: none, 1: iter info, 2: diagnostic info
      cplex.setParam(IloCplex::RootAlg, solverAlgo);
      cplex.setParam(IloCplex::BarDisplay, solverVerbose);
      cplex.setParam(IloCplex::SimDisplay, solverVerbose);

      int nIters[MAX_ITERS_1];
      double *w_iter1 = new double[wSz];
      double *w_iter2 = new double[wSz];
      double **objValues = buildMatrix<double>(MAX_ITERS_1, MAX_ITERS_2);
      double **wDeviation = buildMatrix<double>(MAX_ITERS_1, MAX_ITERS_2);

      int curAllWsMem = 10; // current allocation memory for all_ws;
      double *all_ws = (double*) mxCalloc(wSz*curAllWsMem, sizeof(double));


      double *posFeats   = new double[nBin*nTr];
      double *negFeats   = new double[nBin*nTr];
      int *negClassIdxs  = new int[nTr];
      double *negSvmVals = new double[nTr];
      double *posSvmVals = new double[nTr];
      double *posFeats_ptr, *negFeats_ptr;

      featGen.getInitFeats("pos", posFeats);

      if (VERBOSE >= 2){
         mexPrintf("   Init PosFeats:\n");
         for (int j = 0; j < nBin; j++){
            mexPrintf("   ");
            for (int i = 0; i < nTr; i++) mexPrintf("%.3f ", posFeats[i*nBin + j]);
            mexPrintf("\n");
         }
      }


      int nIter1 = 0;
      for (int iter1=0; iter1 < MAX_ITERS_1; iter1++){// Coordinate descent
         nIter1++;
         IloConstraintArray constrs(env);
         int nConstr = 0;

         mexPrintf("Outer: %d, adding init. constrs\n", iter1 + 1);
         // Generate initial negative constraints
         for(int i = 0; i < INIT_CONSTR_PER_SAMPLE; i++){
            featGen.getInitFeats("neg", negFeats, negClassIdxs);
            posFeats_ptr = posFeats;
            negFeats_ptr = negFeats;

            for (int j=0; j < nTr; j++){
               int trueClassOffset = lbs[j]*nBin; // true label
               int vioClassOffset  = negClassIdxs[j]*nBin; // violating class

               IloExpr constrExp(env);
               for (int k=0; k <nBin; k++){
                  constrExp += (*posFeats_ptr++)*xw[trueClassOffset++] - (*negFeats_ptr++)*xw[vioClassOffset++];
               }
               constrs.add(constrExp >= 1 - xxi[j]);
               mod.add(constrs[nConstr++]);
               constrExp.end();
            }
            if (VERBOSE >= 3){
               mexPrintf("   Init NegFeats:\n   ");
               for (int i=0; i < nTr; i++) mexPrintf("%5d ", negClassIdxs[i]);
               mexPrintf("\n");
               for (int j = 0; j < nBin; j++){
                  mexPrintf("   ");
                  for (int i = 0; i < nTr; i++) mexPrintf("%5.3f ", negFeats[i*nBin + j]);
                  mexPrintf("\n");
               }
            }
         }

         double curObjVal = 0;
         for(int iter2 = 0;iter2 < MAX_ITERS_2; iter2++){//inner loop, constraint generation
            nIters[iter1] = iter2+1;
            mexPrintf("%s, C: %g, outer iter: %d, inner iter: %d\n", runIDString, C, iter1 + 1, iter2 + 1);

            // Optimize the problem and obtain solution.
            mexPrintf("   Cplex: optimizing... "); startTime = clock();
            startTime = clock();
            if ( !cplex.solve() ) {
               env.error() << "Failed to optimize QP" << endl;
               throw(-1);
            }

            curObjVal = cplex.getObjValue();
            objValues[iter1][iter2] = curObjVal;
            env.out() << "status: " << cplex.getStatus() <<
               ", elapse time: " << diffclock(clock(), startTime) << endl;

            cplex.getValues(w, xw);
            cplex.getValues(xi, xxi);

            bool isDevEnough = isDeviateEnough(w, w_iter2, &deviation);
            wDeviation[iter1][iter2] = deviation;

            if (VERBOSE >= 3){
               env.out() << "   w  = " << w  << endl;
               env.out() << "   xi = " << xi << endl;
            }
            if (VERBOSE >= 1){
               double w_L2 = 0;
               for (int i=0; i < wSz; i++) w_L2 += w[i]*w[i];
               double xi_L1 = 0;
               for (int i=0; i < nTr; i++) xi_L1 += xi[i];
               mexPrintf("   (1/nClass)*w*w: %g, C/nTr*xi: %g, obj.val: %g\n",
                     (1.0f/nClass)*w_L2, (C/nTr)*xi_L1, curObjVal);
               mexPrintf("   w deviation: %g\n", deviation);
            }

            if (!isDevEnough){
               mexPrintf("   deviation is small, quitting to the outer loop,  final obj. val: %g\n", curObjVal);
               break;
            }


            //find maximum violated constraints
            mexPrintf("   Generating new constraints... ");
            for(int i = 0; i<wSz; i++)  w_iter2[i] = w[i]; // remember old weights

            featGen.getSegFeats(w, "neg", negFeats, negClassIdxs);

            if (VERBOSE >= 3){
               mexPrintf("   NegFeats:\n   ");
               for (int i=0; i < nTr; i++) mexPrintf("%5d ", negClassIdxs[i]);
               mexPrintf("\n");
               for (int j = 0; j < nBin; j++){
                  mexPrintf("   ");
                  for (int i = 0; i < nTr; i++) mexPrintf("%5.3f ", negFeats[i*nBin + j]);
                  mexPrintf("\n");
               }
            }


            cmpMCSVMVals(negFeats, negClassIdxs, w_iter2, nTr, nBin, negSvmVals);
            cmpMCSVMVals(posFeats, lbs, w_iter2, nTr, nBin, posSvmVals);

            int nNewConstr = 0;
            double extraConstrVio = 0;

            for(int j = 0;j < nTr;j++){
               double aux = posSvmVals[j] - negSvmVals[j] - 1 + xi[j];
               if (aux < 0){
                  extraConstrVio -= aux;
                  int trueClassOffset = lbs[j]*nBin; // true label
                  int vioClassOffset  = negClassIdxs[j]*nBin; // violating class
                  posFeats_ptr = posFeats + j*nBin;
                  negFeats_ptr = negFeats + j*nBin;

                  IloExpr constrExp(env);
                  for (int k=0; k <nBin; k++){
                     constrExp += (*posFeats_ptr++)*xw[trueClassOffset++] - (*negFeats_ptr++)*xw[vioClassOffset++];
                  }
                  constrs.add(constrExp >= 1 - xxi[j]);
                  constrExp.end();
                  mod.add(constrs[nConstr++]);
                  nNewConstr++;
               }
            }
            extraConstrVio *= C/nTr;
            mexPrintf("new/total: %d/%d, elapsed time: %.2f, extra violation: %f\n",
                  nNewConstr, nConstr, diffclock(clock(), startTime), extraConstrVio);
            if ((!nNewConstr)||(extraConstrVio < EXTRA_VIOLATION_TOLFAC*curObjVal)) {
               mexPrintf("   extraConstrVio is small, exit to outer loop, obj. val: %g\n", curObjVal);
               break;
            }
         }

         if (iter1 >= curAllWsMem) {
            curAllWsMem *= 2; // double the allocation
            all_ws = (double*) mxRealloc(all_ws, curAllWsMem*wSz*sizeof(double));
         }
         double *all_ws_ptr = all_ws + iter1*wSz;
         for (int t =0; t < wSz; t++) *all_ws_ptr++ = w[t];


         if (!isDeviateEnough(w, w_iter1, &deviation)){
            mexPrintf("===>outer loop, w deviation: %g, deviation is small, quitting\n", deviation);
            break;
         } else if ((iter1 > 0)  &&
               (objValues[iter1][nIters[iter1]-1] >
               (1- OBJ_REDUCTION_TOLFAC)*objValues[iter1-1][nIters[iter1-1]-1])) {
            mexPrintf("===>outer loop, Obj.value does not deviate much, quitting\n");
            break;
         }


         /* Modify the positive constraints*/
         mexPrintf("---------------------------------------------------------"
                   "\nFinding new fg segments...");
         startTime = clock();
         for(int i = 0;i < wSz;i++)  w_iter1[i] = w[i]; // remember old weights
         featGen.getSegFeats(w, "pos", posFeats);
         mexPrintf("   elapsed time: %.2f\n", diffclock(clock(), startTime));

         if (VERBOSE >= 2){
            mexPrintf("   current Ws:\n");
            for (int j=0; j < nBin; j++){
               mexPrintf("   ");
               for (int i = 0; i < nClass; i++) mexPrintf("%.5f ", w[i*nBin + j]);
               mexPrintf("\n");
            }
            mexPrintf("   PosFeats:\n");
            for (int j = 0; j < nBin; j++){
               mexPrintf("   ");
               for (int i = 0; i < nTr; i++) mexPrintf("%.3f ", posFeats[i*nBin + j]);
               mexPrintf("\n");
            }
         }


         // Clear xi, if not, no more negative constraint is generated
         // w need not be cleared
         for (int i = 0; i < nTr; i++) xi[i] = 0;

         // Remove old negative constraints
         for (int i=0; i < nConstr; i++) mod.remove(constrs[i]);
         constrs.endElements();

      }

      mexPrintf("==========================================\n");
      mexPrintf("m_SegMCSVM_CCCP: done, some statistics ...\n");
      mexPrintf("   # outer iters: %d, # inner iters:", nIter1);
      for (int i=0; i < nIter1; i++) mexPrintf("%d ", nIters[i]);
      mexPrintf("\n");

      /* Create outputs*/
      optRslt.w = new double[wSz];
      for (int i=0; i < wSz; i++) optRslt.w[i] = w[i];
      optRslt.nIter1 = nIter1;
      optRslt.nCGIters = new int[nIter1];
      for (int i=0; i < nIter1; i++) optRslt.nCGIters[i] = nIters[i];
      optRslt.objVals = new double*[nIter1];
      optRslt.wDeviation = new double*[nIter1];
      for (int i=0; i < nIter1; i++){
         int nIter2 = nIters[i];
         optRslt.objVals[i] = new double[nIter2];
         optRslt.wDeviation[i] = new double[nIter2];
         for (int j=0; j < nIter2; j++){
            optRslt.objVals[i][j] = objValues[i][j];
            optRslt.wDeviation[i][j] = wDeviation[i][j];
         }
      }

      optRslt.all_ws = new double[wSz*nIter1*sizeof(double)];
      memcpy(optRslt.all_ws, all_ws, wSz*nIter1*sizeof(double));


      //free memory
      delete [] w_iter1;
      delete [] w_iter2;
      delete [] negFeats;
      delete [] posFeats;
      delete [] negClassIdxs;
      delete [] posSvmVals;
      delete [] negSvmVals;

      mxFree(all_ws);
      destroyMatrix<double>(wDeviation, MAX_ITERS_1, MAX_ITERS_2);
      destroyMatrix<double>(objValues, MAX_ITERS_1, MAX_ITERS_2);
   } catch (IloException& e) {
      cerr << "m_SegMCSVM_CCCP: concert exception caught: " << e << endl;
   } catch (exception& e) {
      cerr << "m_SegMCSVM_CCCP: standard exception caught: " << e.what() << endl;
   } catch (...) {
      cerr << "m_SegMCSVM_CCCP: unknown exception caught." << endl;
   }

   env.end();
   return optRslt;
}
#endif
