function Segs = ml_findMulti1DSegs(kpAssign, imW, binW, nSeg)
% function ml_findMultiSegs(histo, imW, binW, nSeg, szRng, thres)
% Finding multiple 1D segments from the 1D histograms.
% The feature vector representing a group of segments is a histogram, hist,
% which is the addition of all histograms. The objective is to minimize 
% sum_i {binW(i)*hist(i)}
% This function find GLOBAL minimum!
% Inputs:
%   kpAssign: a n*2 matrix for histogram bin assignments of keypoints. 
%       kpAssign(i,1), kpAssign(i,2) are the col#, bin#. 
%       1 <= col# <= imW; 1 <= bin# <= k, where k
%       is the length of binWeights.
%   imW: width or length of 1D signal.
%   binW: a k*1 vector of weights for histogram bins, weights can be
%       negative
%   nSeg: number of segments to extract, must a pos integer. The default is 1. 
% Output:
%   Segs: a 3*K matrix with K <= nSeg. Segs([1,2],i) marks the left and
%   right ends of the interval. Segs(3,i) is the value of the interval.
% Examples:
%   signal_1D = ones(1,40);
%   signal_1D([4:7, 10:15, 25:35]) = 2;
%   n = length(signal_1D);
%   kpAssign = [(1:n)', signal_1D(:)];
%   ml_findMulti1DSegs(kpAssign, n, [1; -1], 3)
%   ml_findMulti1DSegs(kpAssign, n, [1; -1], 2)
%   ml_findMulti1DSegs(kpAssign, n, [3; -1], 2)
%   
% See also:
%   ml_findMulti1DSegs_seq.m: this does not guarantee to find global
%   minimum. However, it can incorporate size constraint.
%   
% By: Minh Hoai Nguyen (minhhoai@gmail.com)
% Date: 13 Jan 2009
% Last modified: 31 Jan 2009


if size(binW, 2) ~= 1
    error('binWeights should be a column vector');
end;

nHistBin = size(binW, 1);

% Check kpAssign
if size(kpAssign, 2) ~= 2
    error('kpAssign must be n*2 matrix');
end;

if sum(kpAssign(:) <= 0) > 0
    error('entries of kpAssign should be positive integers');
end

if sum(kpAssign(:,1) > imW) > 0
    error('entries of kpAssign(:,1) must be <= imW');
end;
if sum(kpAssign(:,2) > nHistBin) > 0
    error('entries of kpAssign(:,2) must be smaller than the length of binWeights');
end

if (nSeg < 1)
    error('Number of segment requested must be at least 1');
end;

Segs = m_mexFindMulti1DSegs(kpAssign, imW, binW, nSeg);



